/*
 * gSXprog by Eric Smith, based on sxprog by Loren Blaney
 *
 * Copyright 2001-2003 Eric Smith <eric@brouhaha.com>
 * Rev F SX-Key support added by Jim Garlick <garlick@llnl.gov>
 *
 * $Id: sxkey_new.c,v 1.4 2003/11/18 13:59:42 garlick Exp $
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>            /* to get speed (B19200 etc.) */
#include <assert.h>
#include <libgen.h>
#include <getopt.h>

#include "gsasm.h"
#include "sxkey.h"
#include "hex_file.h"
#include "sxkey_new.h"

static void new_fini(sxkey_t * key);
static boolean _reconnect(sxkey_t * key);

int debug_level = 0;

/*
 * Read a byte from SX-Key.  If 0, no error.  If nonzero, 
 * display error string on stderr.
 */
static boolean _check_error(int fd, int ms_timeout)
{
    boolean ret = false;
    u8 c;
    
    if (!sx_get_byte(fd, ms_timeout, &c)) {
        if (debug_level >= 2)
            fprintf(stderr, "GSASM: check_error timeout\n");
        return false;
    }
    switch (c) {
    case 0:                    /* all OK */
        ret = true;
        break;
    case 1:
        fprintf(stderr, "GSASM: Vpp generation failed\n");
        break;
    case 2:
        fprintf(stderr, "GSASM: chip connection failed\n");
        break;
    case 3:
        fprintf(stderr, "GSASM: programming failed\n");
        break;
    default:
        if (debug_level >= 2)
            fprintf(stderr, "GSASM: comm failed (%02x)\n", c);
        break;
    }
    return ret;
}

static boolean new_reset(sxkey_t * key)
{
    if (!sx_send_byte(key->fd, SXKEY_NEW_CMD_RESET))        /* Reset */
        return false;
    if (!_check_error(key->fd, SXKEY_NEW_TIMEOUT))          /* Check error */
        return false;
    return true;
}

/* 
 * ICD2053B Cypress Programable Clock Generator
 * 
 * f(vco) = 2 * f(ref) * P / Q ; 50MHz <= f(vco) <= 150MHz
 * f(out) = f(vco) / 2**Mux
 * P' = P - 3 ; Q' = Q - 2
 * program word: P'(7bits):Duty(1bit):Mux(3bits):Q'(7bits):I(4bits)
 * 
 * The bit stuffing required by the ICD2053B is done by the 
 * on board SX20AC micro
 * 
 * SX-Key clock setting code from Emil <emil@la.mine.nu>.
 */
static int _icd2053b( float freq )
{
    int           p, q, div, err, minp, minq, minerr;
    float         fvco, pq, truefreq;
    int           pgmword = 0;

    if (freq > MAX_CLOCK )
        return 0x010020;        /* 50MHz */
    for (div = 0; div < 8; div++)
        if ((fvco = freq * (1 << div)) >= 50.)
            break;
    if(div == 8)
        return 0x010020;        /* 50MHz */
    pq = fvco / REF_CLOCK / 2;
    minerr = 100;
    minp = 3;
    minq = 3;
    for (q = 3; q <= 129; q++) {
        p = pq * q + .5;
        if (p > 129)
            break;
        err = (pq * q - p) * 1000;
        if (abs( err ) < minerr) {
            minerr = abs( err );
            minp = p;
            minq = q;
        }
    }
    truefreq = 2 * REF_CLOCK * minp / minq / (1 << div);
    pgmword |= (minp - 3) << 15;
    pgmword |= div << 11;
    pgmword |= (minq - 2) << 4;
    if (fvco >= 80.)
        pgmword |= 8;
    if (debug_level > 0) {
        fprintf(stderr, "Fosc: %f\n", fvco );
        fprintf(stderr, "P/Q: %f\n", pq );
        fprintf(stderr, "P: %d Q: %d\n", minp, minq );
        fprintf(stderr,  "Word: 0x%06x\n", pgmword );
    }
    fprintf(stderr, "GSASM: frequency %.3f MHz\n", truefreq);
    return pgmword;
}

static boolean new_clock_off(sxkey_t * key)
{
    int fd = key->fd;
    boolean r;
    
    r = sx_send_byte(fd, SXKEY_NEW_CMD_CLOCK_DISABLE);
    if (!r)
        fprintf(stderr, "GSASM: clock_off failed\n");
    return r;
}

boolean new_clock_on(sxkey_t * key, float frequency)
{
    int pgmword;

    /* This has been verified for SXKey Rev.E and Rev.F */
    pgmword = _icd2053b( frequency );
    if (!sx_send_byte(key->fd, SXKEY_NEW_CMD_CLOCK_ENABLE)) /* Enable clk */
        return false;
    if (!sx_send_byte(key->fd, pgmword & 0xff))             /* Freq low byte */
        return false;
    if (!sx_send_byte(key->fd, (pgmword >> 8) & 0xff))      /* Freq mid byte */
        return false;
    if (!sx_send_byte(key->fd, (pgmword >> 16) & 0xff))     /* Freq hi byte */
        return false;
    if (!_check_error(key->fd, SXKEY_NEW_TIMEOUT))          /* Check error */
        return false;
    return true;
}

static boolean _read_prolog(sxkey_t * key)
{
    if (!sx_send_byte(key->fd, SXKEY_NEW_CMD_READ))     /* Send 'read' cmd */
        return false;
    if (!_check_error(key->fd, SXKEY_NEW_TIMEOUT))      /* Check error */
        return false;
    return true;
}

static boolean _read_fuses(sxkey_t * key, sx_t * sx)
{
    if (!sx_send_byte(key->fd, 3))                      /* Read to read 3 */
        return false;
    if (!sx_get_word(key->fd, SXKEY_NEW_TIMEOUT, &sx->device))
        return false;                                   /* Read DEVICE */
    if (!sx_get_word(key->fd, SXKEY_NEW_TIMEOUT, &sx->rom[FUSEX_LOC]))
        return false;                                   /* Read FUSEX */
    if (!sx_get_word(key->fd, SXKEY_NEW_TIMEOUT, &sx->rom[FUSE_LOC]))
        return false;                                   /* Read FUSE */
    if (!_check_error(key->fd, SXKEY_NEW_TIMEOUT))
        return false;
    return true;
}

static boolean _read_code(sxkey_t * key, sx_t * sx)
{
    int size_rom = sx_rom_size(sx);     /* assumes fuses have been read */
    int i, j;

    assert(size_rom % 0x80 == 0);
    for (i = 0; i < size_rom; i += 0x80) {              /* Read CODE */
        if (!sx_send_byte(key->fd, 0x80))               /* Ready to read 0x80 */
            return false;
        for (j = 0; j < 0x80; j++) {                    /* Read 0x80 words */
            if (!sx_get_word(key->fd, SXKEY_NEW_TIMEOUT, &sx->rom[i + j]))
                return false;
        }
        if (!_check_error(key->fd, SXKEY_NEW_TIMEOUT))
            return false;                               /* Check error */
    }
    return true;
}

static boolean _read_id(sxkey_t * key, sx_t * sx)
{
    int i;

    if (!sx_send_byte(key->fd, SX_ID_SIZE))             /* Read to read ... */
        return false;
    for (i = 0; i < SX_ID_SIZE; i++) {                  /* Read Id */
        if (!sx_get_word(key->fd, SXKEY_NEW_TIMEOUT, &sx->rom[ID_LOC + i]))
            return false;
    }
    if (!_check_error(key->fd, SXKEY_NEW_TIMEOUT))
        return false;                                   /* Check error */
    return true;
}

static boolean _read_epilog(sxkey_t * key)
{
    if (!sx_send_byte(key->fd, 0))                      /* End read */
        return false;
    if (!sx_get_word(key->fd, SXKEY_NEW_TIMEOUT, NULL))
        return false;                                   /* Skip timing word */
    if (!sx_get_byte(key->fd, SXKEY_NEW_TIMEOUT, NULL))
        return false;                                   /* Skip 0 */
    if (!new_reset(key))                                /* Reset */
        return false;
    return true;
}

static boolean new_read(sxkey_t * key, sx_t * sx)
{
    fprintf(stderr, "GSASM: reading chip...");
    if (!_read_prolog(key))
        goto err;
    if (!_read_fuses(key, sx))
        goto err;
    if (!_read_code(key, sx))   /* assumes fuses have been read */
        goto err;
    if (!_read_id(key, sx))
        goto err;
    if (!_read_epilog(key))
        goto err;
    fprintf(stderr, "OK\n");
    return true;
  err:
    fprintf(stderr, "failed\n");
    return false;
}

static boolean new_read_fuses(sxkey_t * key, sx_t * sx)
{
    if (!_read_prolog(key))
        return false;
    if (!_read_fuses(key, sx))
        return false;
    if (!_read_epilog(key))
        return false;
    return true;
}

static boolean _erase(sxkey_t * key)
{
    u16 sum_time = SX_NEW_ERASE_MS;
    u16 dose_time;

    while (sum_time > 0) {
        dose_time = (sum_time > 255) ? 255 : sum_time;
        sum_time -= dose_time;

        if (!sx_send_byte(key->fd, SXKEY_NEW_CMD_ERASE))
            goto err;                                   /* Send erase command */
        if (!sx_send_byte(key->fd, dose_time))          /* Send erase time */
            goto err;
        if (!_check_error(key->fd, SX_NEW_ERASE_MS + dose_time))
            goto err;                                   /* Check error */
    }
    return true;
  err:
    return false;
}

static boolean new_erase(sxkey_t * key)
{
    boolean r;

    fprintf(stderr, "GSASM: erasing chip...");
    if ((r = _erase(key)))
        fprintf(stderr, "OK\n");
    else
        fprintf(stderr, "failed\n");
    return r;
}

static boolean _prog_prolog(sxkey_t *key)
{
    if (!sx_send_byte(key->fd, SXKEY_NEW_CMD_PROGRAM))  /* Program device */
        return false;
    if (!_check_error(key->fd, SXKEY_NEW_TIMEOUT))
        return false;                                   /* Check error */
    return true;
}

static boolean _prog_epilog(sxkey_t *key, u16 *freq)
{
    if (!sx_send_byte(key->fd, 0))                      /* End programming */
        return false;
    if (!sx_get_word(key->fd, SXKEY_NEW_TIMEOUT, freq)) /* timing word */
        return false;
    if (!sx_get_byte(key->fd, SXKEY_NEW_TIMEOUT, NULL)) /* 0 */
        return false;
    return true;
}

static boolean _prog_fusex(sxkey_t *key, u16 fusex, u16 *freq)
{
    if (!_prog_prolog(key))
        return false;
    if (!sx_send_byte(key->fd, 1))                      /* Prog 1 word */
        return false;
    if (!sx_send_byte(key->fd, SX_NEW_FUSE_PROG_MS))    /* Send program time */
        return false;
    if (!sx_send_word(key->fd, fusex))
        return false;                                   /* Program FUSEX */
    if (!_check_error(key->fd, SXKEY_NEW_PROG_FUSE_TIMEOUT(1)))
        return false;                                   /* Check error */
    if (!_prog_epilog(key, freq))
        return false;
    return true;
}

static boolean _prog_verify(sxkey_t *key,  sx_t *sx, boolean idskip)
{                                                   
    sx_t sx_r;

    if (!new_read(key, &sx_r))
        return false;
    if (!sx_compare(sx, &sx_r, false, idskip))
        return false;
    return true;
}

/* 
 * This function is called during programming to determine optimal
 * values for FUSEX bits used for trimming the SX device's internal
 * oscillator.  The result stored in fusex should be ANDed with 
 * the fusex value read in from the hex file.
 *
 * ATTRIB: borrowed algorithm which minimizes erasures from open source 
 * sx-key pascal source code.
 */

#define ABS(i)  ((i) < 0 ? -(i) : (i))
typedef struct { 
    u16 fusex; 
    u16 freq; 
} trim_t;
#define SXKEY_CLK_HZ_TARGET     4000000
#define SXKEY_CLK_HZ(rsp)       (1266035968L/(rsp))

static boolean _trim_oscillator(sxkey_t *key, u16 *fusex)
{
    u8 i;
    u32 error, min_error;
    trim_t trim[8] = { 
        { .fusex = 0xfff }, /* 0 */
        { .fusex = 0x7ff },
        { .fusex = 0x5ff },
        { .fusex = 0x4ff }, 
        { .fusex = 0xdff }, /* 4 */
        { .fusex = 0xcff },
        { .fusex = 0xeff }, /* 6 */
        { .fusex = 0x6ff },
    };

    fprintf(stderr, "GSASM: trimming internal SX oscillator");

    min_error = ~0;
    for (i = 0; i < 8; i++) {
        if(i==0 || i==4 || i==6) 
            if(!_erase(key))                            /* Erase */
                goto err;
        if (!_prog_fusex(key, trim[i].fusex, &trim[i].freq))
            goto err;
        if (debug_level > 0) {
            fprintf(stderr, "\nGSASM: trim bits 0x%x yield %ld Hz", trim[i].fusex, SXKEY_CLK_HZ(trim[i].freq));
        } else
            fprintf(stderr, "."); 
        fflush(stderr);

        error = ABS(SXKEY_CLK_HZ_TARGET - SXKEY_CLK_HZ(trim[i].freq));
        if (error < min_error) {
            min_error = error;
            *fusex = trim[i].fusex;
        }
    }

    if (debug_level > 0)
        fprintf(stderr, "\nGSASM: using trim bits 0x%x\n", *fusex);
    else
        fprintf(stderr, "OK\n");
    return true;
err:
    fprintf(stderr, "failed\n");
    return false;
}

/* 
 * Determine how many bytes need to be programmed by looking at
 * the fuse bytes in the image to be programmed.  If --idskip
 * was specified, we can further reduce the size by lopping off 
 * all the bytes at the end of the rom that are 0xfff.
 */
static void _prog_size(sx_t *sx, int *size, boolean idskip)
{
    int i, newsize;

    newsize = sx_rom_size(sx);

    for (i = newsize; i < SX_MAX_ROM_SIZE; i++)
        if (sx->rom[i] != 0xfff) {
            fprintf(stderr, 
                    "GSASM: warning: device configured for ROM size %d,\n"
                    "GSASM:          but hex file has data at address %d\n", newsize, i);
            break;
        }

    if (idskip) {
        for (i = newsize - 1; i >= 0; i--) {
            if (sx->rom[i] != 0xfff)
                break;
        }    
        newsize = (i + 1);
	
        if (newsize % 0x20)
            newsize += 0x20 - (newsize % 0x20);
    }
    *size = newsize;
}

static boolean new_prog(sxkey_t * key, sx_t * sx, boolean trim, boolean idskip)
{
    int i, j;
    int size_rom;
    int fd = key->fd;

    _prog_size(sx, &size_rom, idskip);                  /* Find prog size */

    if (key->fw_rev != SXKEY_BLITZ) {
        if (!new_clock_off(key))                        /* Clock off */
            goto err;
    }

    if (trim) {                                         /* Trim IRC */
        u16 fusex_trim;

        sx->rom[FUSEX_LOC] |= (IRCTRIM0 | IRCTRIM1 | IRCTRIM2);
        _trim_oscillator(key, &fusex_trim);
        sx->rom[FUSEX_LOC] &= fusex_trim;
    }

    fprintf(stderr, "GSASM: erasing chip...");
    if (!_erase(key))                                   /* Erase */
        goto err;
    fprintf(stderr, "OK\n");

    if (!_prog_prolog(key))
        goto err;

    if (debug_level > 0)
        sx_show_fuse_bits(sx);

    fprintf(stderr, "GSASM: programming FUSE/FUSEX...");
    if (!sx_send_byte(fd, 2))                           /* Prog 2 words */
        goto err;
    if (!sx_send_byte(fd, SX_NEW_FUSE_PROG_MS))         /* Send program time */
        goto err;
    if (!sx_send_word(fd, sx->rom[FUSEX_LOC]))
        goto err;                                       /* Program FUSEX */
    if (!sx_send_word(fd, sx->rom[FUSE_LOC]))           /* Program FUSE */
        goto err;
    if (!_check_error(fd, SXKEY_NEW_PROG_FUSE_TIMEOUT(2)))
        goto err;                                       /* Check error */
    fprintf(stderr, "OK\n");

    fprintf(stderr, "GSASM: programming ROM (%d words)...",
            size_rom);
    assert(size_rom % 0x20 == 0);
    for (i = 0; i < size_rom; i += 0x20) {
        fprintf(stderr, "%-5d\b\b\b\b\b", i); fflush(stderr);
        if (!sx_send_byte(fd, 0x20))                    /* Program 32 words */
            goto err;
        if (!sx_send_byte(fd, SX_NEW_CODE_PROG_MS))
            goto err;                                   /* Send program time */
        for (j = 0; j < 0x20; j++) {
            if (!sx_send_word(fd, sx->rom[i + j]))
                goto err;                               /* Program CODE */
        }
        if (!_check_error(fd, SXKEY_NEW_PROG_CODE_TIMEOUT(0x20)))
            goto err;                                   /* Check error */
        fputc( '.', stderr );
        fflush( stderr );
    }
    fprintf(stderr, "OK   \n");

    if (!idskip) {
        fprintf(stderr, "GSASM: programming ID (%d words)...",
                SX_ID_SIZE);
        if (!sx_send_byte(fd, SX_ID_SIZE))              /* Prog <size> words */
            goto err;
        if (!sx_send_byte(fd, SX_NEW_CODE_PROG_MS))     /* Send program time */
            goto err;
        for (i = 0; i < SX_ID_SIZE; i++) {              /* Program ID */
            if (!sx_send_word(fd, sx->rom[ID_LOC + i]))
                goto err;
        }
        if (!_check_error(fd, SXKEY_NEW_PROG_CODE_TIMEOUT(SX_ID_SIZE)))
            goto err;
        fprintf(stderr, "OK\n");
    }

    if (!_prog_epilog(key, NULL))
        goto err;
    
    if (!new_reset(key))                                /* Reset */
        goto err;

    if (!_prog_verify(key, sx, idskip)) {               /* Verification */
        fprintf(stderr, "GSASM: verification failed\n");
        goto err;
    } else
        fprintf(stderr, "GSASM: verification succeded\n");

#if 0
    /* XXX could turn clock back on here but since we don't know whether
     * it was on to begin with and at what frequency, perhaps it's better to 
     * leave it off and put a note in the man page about the clock being 
     * disabled as a side-effect of programming.
     */
    switch( key->fw_rev ) {
        case SXKEY_REV_E:
        case SXKEY_REV_F:
            if (!new_clock_on(key, 50.))
                goto err;
            break;
    }
#endif
    return true;
  err:
    fprintf(stderr, "failed\n");
    return false;
}

/*
 * Close serial port and log off SX-Key.
 */
static void new_fini(sxkey_t * key)
{
    if (!sx_send_byte(key->fd, '\0'))
        fprintf(stderr, "GSASM: logoff failed\n");
    sx_serial_close(key->fd);
    key->fd = -1;
    sx_unsetkeys();
}

static boolean _reconnect(sxkey_t * key)
{
    u8 c1, c2;

    sx_unsetkeys();

    if (!sx_send_byte(key->fd, '\0'))
        goto err;
    if (!sx_serial_flush(key->fd))
        goto err;
    if (!sx_send_str(key->fd, " SX-Key"))
        goto err;
    if (!sx_expect_str(key->fd, SXKEY_NEW_TIMEOUT, "SX-Key"))
        goto err;
    if (!sx_get_byte(key->fd, SXKEY_NEW_TIMEOUT, &c1))
        goto err;
    if (!sx_get_byte(key->fd, SXKEY_NEW_TIMEOUT, &c2))
        goto err;
    sx_setkeys(c1, c2);
    if (!sx_get_byte(key->fd, SXKEY_NEW_TIMEOUT, &key->fw_rev))
        goto err;
    return true;
  err:
    return false;
}

/*
 * Open serial port and log on to the SX-Key device.
 */
boolean new_init(char *dev, speed_t baud, sxkey_t * key, sx_t * sx)
{
    if (!sx_serial_open(dev, baud, &key->fd))
        return false;

    if (!_reconnect(key))                               /* sets key->fw_rev */
        return false;

    key->cmd_prog = new_prog;
    key->cmd_erase = new_erase;
    key->cmd_read = new_read;
    key->cmd_read_fuses = new_read_fuses;
    key->cmd_clock_on = new_clock_on;
    key->cmd_clock_off = new_clock_off;
    key->cmd_reset = new_reset;
    key->fini = new_fini;
    key->speed = baud;
    key->device = dev;

    switch (key->fw_rev) {
        case SXKEY_REV_E:
            fprintf(stderr, "GSASM: SX-Key Rev E");
            break;
        case SXKEY_REV_F:
            fprintf(stderr, "GSASM: SX-Key Rev F");
            break;
        case SXKEY_BLITZ:
            fprintf(stderr, "GSASM: SX-Blitz");
            break;
        default:
            fprintf(stderr, "GSASM: SX-Key Rev %d", key->fw_rev);
            break;
    }

    if (!new_read_fuses(key, sx))             /* get device/fusex */
        return false;
    if (SX_28_MODE(sx->device))
        fprintf(stderr, " with SX 18/28 type device\n");
    else if (SX_52_MODE(sx->device))
        fprintf(stderr, " with SX 48/52 type device\n");
    else {
        fprintf(stderr, " with SX ??? (device word: 0x%x)\n", sx->device);
        return false;
    }

    return true;
}

/*
 * em:Local Variables: 
 * em:indent-tabs-mode:nil
 * em:c-basic-offset:4
 * em:End:
 * vi:tabstop=4 shiftwidth=4 expandtab
 */
